import argparse
import logmuse
import os
import sys
import yaml

from eido import inspect_project
from pephubclient import PEPHubClient
from typing import Tuple, List
from ubiquerg import VersionInHelpParser

from . import __version__
from .const import *
from .divvy import DEFAULT_COMPUTE_RESOURCES_NAME, select_divvy_config
from .exceptions import *
from .looper import *
from .parser_types import *
from .project import Project, ProjectContext
from .utils import (
    dotfile_path,
    enrich_args_via_cfg,
    is_registry_path,
    read_looper_dotfile,
    read_looper_config_file,
    read_yaml_file,
    initiate_looper_config,
    init_generic_pipeline,
)


class _StoreBoolActionType(argparse.Action):
    """
    Enables the storage of a boolean const and custom type definition needed
    for systematic html interface generation. To get the _StoreTrueAction
    output use default=False in the add_argument function
    and default=True to get _StoreFalseAction output.
    """

    def __init__(self, option_strings, dest, type, default, required=False, help=None):
        super(_StoreBoolActionType, self).__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=0,
            const=not default,
            default=default,
            type=type,
            required=required,
            help=help,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, self.const)


def build_parser():
    """
    Building argument parser.

    :return argparse.ArgumentParser
    """
    # Main looper program help text messages
    banner = "%(prog)s - A project job submission engine and project manager."
    additional_description = (
        "For subcommand-specific options, " "type: '%(prog)s <subcommand> -h'"
    )
    additional_description += "\nhttps://github.com/pepkit/looper"

    parser = VersionInHelpParser(
        prog="looper",
        description=banner,
        epilog=additional_description,
        version=__version__,
    )

    aux_parser = VersionInHelpParser(
        prog="looper",
        description=banner,
        epilog=additional_description,
        version=__version__,
    )
    result = []
    for parser in [parser, aux_parser]:
        # Logging control
        parser.add_argument(
            "--logfile",
            help="Optional output file for looper logs " "(default: %(default)s)",
        )
        parser.add_argument("--logging-level", help=argparse.SUPPRESS)
        parser.add_argument(
            "--dbg",
            action="store_true",
            help="Turn on debug mode (default: %(default)s)",
        )

        parser = logmuse.add_logging_options(parser)
        subparsers = parser.add_subparsers(dest="command")

        def add_subparser(cmd):
            message = MESSAGE_BY_SUBCOMMAND[cmd]
            return subparsers.add_parser(
                cmd,
                description=message,
                help=message,
                formatter_class=lambda prog: argparse.HelpFormatter(
                    prog, max_help_position=37, width=90
                ),
            )

        # Run and rerun command
        run_subparser = add_subparser("run")
        rerun_subparser = add_subparser("rerun")
        collate_subparser = add_subparser("runp")
        table_subparser = add_subparser("table")
        report_subparser = add_subparser("report")
        destroy_subparser = add_subparser("destroy")
        check_subparser = add_subparser("check")
        clean_subparser = add_subparser("clean")
        inspect_subparser = add_subparser("inspect")
        init_subparser = add_subparser("init")
        init_piface = add_subparser("init-piface")
        link_subparser = add_subparser("link")

        # Flag arguments
        ####################################################################
        for subparser in [run_subparser, rerun_subparser, collate_subparser]:
            subparser.add_argument(
                "-i",
                "--ignore-flags",
                default=False,
                action=_StoreBoolActionType,
                type=html_checkbox(checked=False),
                help="Ignore run status flags? Default=False",
            )

        for subparser in [
            run_subparser,
            rerun_subparser,
            destroy_subparser,
            clean_subparser,
            collate_subparser,
        ]:
            subparser.add_argument(
                "-d",
                "--dry-run",
                action=_StoreBoolActionType,
                default=False,
                type=html_checkbox(checked=False),
                help="Don't actually submit the jobs.  Default=False",
            )

        # Parameter arguments
        ####################################################################
        for subparser in [run_subparser, rerun_subparser, collate_subparser]:
            subparser.add_argument(
                "-t",
                "--time-delay",
                metavar="S",
                type=html_range(min_val=0, max_val=30, value=0),
                default=0,
                help="Time delay in seconds between job submissions",
            )

            subparser.add_argument(
                "-x",
                "--command-extra",
                default="",
                metavar="S",
                help="String to append to every command",
            )
            subparser.add_argument(
                "-y",
                "--command-extra-override",
                metavar="S",
                default="",
                help="Same as command-extra, but overrides values in PEP",
            )
            subparser.add_argument(
                "-f",
                "--skip-file-checks",
                action=_StoreBoolActionType,
                default=False,
                type=html_checkbox(checked=False),
                help="Do not perform input file checks",
            )

            divvy_group = subparser.add_argument_group(
                "divvy arguments", "Configure divvy to change computing settings"
            )
            divvy_group.add_argument(
                "--divvy",
                default=None,
                metavar="DIVCFG",
                help="Path to divvy configuration file. Default=$DIVCFG env "
                "variable. Currently: {}".format(
                    os.getenv("DIVCFG", None) or "not set"
                ),
            )
            divvy_group.add_argument(
                "-p",
                "--package",
                metavar="P",
                help="Name of computing resource package to use",
            )
            divvy_group.add_argument(
                "-s",
                "--settings",
                default="",
                metavar="S",
                help="Path to a YAML settings file with compute settings",
            )
            divvy_group.add_argument(
                "-c",
                "--compute",
                metavar="K",
                nargs="+",
                help="List of key-value pairs (k1=v1)",
            )

        for subparser in [run_subparser, rerun_subparser]:
            subparser.add_argument(
                "-u",
                "--lump",
                default=None,
                metavar="X",
                type=html_range(min_val=0, max_val=100, step=0.1, value=0),
                help="Total input file size (GB) to batch into one job",
            )
            subparser.add_argument(
                "-n",
                "--lumpn",
                default=None,
                metavar="N",
                type=html_range(min_val=1, max_val="num_samples", value=1),
                help="Number of commands to batch into one job",
            )

        check_subparser.add_argument(
            "--describe-codes",
            help="Show status codes description",
            action="store_true",
            default=False,
        )

        check_subparser.add_argument(
            "--itemized",
            help="Show a detailed, by sample statuses",
            action="store_true",
            default=False,
        )

        check_subparser.add_argument(
            "-f",
            "--flags",
            nargs="*",
            default=FLAGS,
            type=html_select(choices=FLAGS),
            metavar="F",
            help="Check on only these flags/status values",
        )

        for subparser in [destroy_subparser, clean_subparser]:
            subparser.add_argument(
                "--force-yes",
                action=_StoreBoolActionType,
                default=False,
                type=html_checkbox(checked=False),
                help="Provide upfront confirmation of destruction intent, "
                "to skip console query.  Default=False",
            )

        init_subparser.add_argument(
            "pep_config", help="Project configuration file (PEP)"
        )

        init_subparser.add_argument(
            "-f", "--force", help="Force overwrite", action="store_true", default=False
        )

        init_subparser.add_argument(
            "-o",
            "--output-dir",
            dest="output_dir",
            metavar="DIR",
            default=None,
            type=str,
        )

        init_subparser.add_argument(
            "-S",
            "--sample-pipeline-interfaces",
            dest=SAMPLE_PL_ARG,
            metavar="YAML",
            default=None,
            nargs="+",
            type=str,
            help="Path to looper sample config file",
        )
        init_subparser.add_argument(
            "-P",
            "--project-pipeline-interfaces",
            dest=PROJECT_PL_ARG,
            metavar="YAML",
            default=None,
            nargs="+",
            type=str,
            help="Path to looper project config file",
        )

        # TODO: add  ouput dir, sample, project pifaces

        init_subparser.add_argument(
            "-p",
            "--piface",
            help="Generates generic pipeline interface",
            action="store_true",
            default=False,
        )

        # Common arguments
        for subparser in [
            run_subparser,
            rerun_subparser,
            table_subparser,
            report_subparser,
            destroy_subparser,
            check_subparser,
            clean_subparser,
            collate_subparser,
            inspect_subparser,
            link_subparser,
        ]:
            subparser.add_argument(
                "config_file",
                nargs="?",
                default=None,
                help="Project configuration file (YAML) or pephub registry path.",
            )
            subparser.add_argument(
                "--looper-config",
                required=False,
                default=None,
                type=str,
                help="Looper configuration file (YAML)",
            )
            # help="Path to the looper config file"
            subparser.add_argument(
                "-S",
                "--sample-pipeline-interfaces",
                dest=SAMPLE_PL_ARG,
                metavar="YAML",
                default=None,
                nargs="+",
                type=str,
                help="Path to looper sample config file",
            )
            subparser.add_argument(
                "-P",
                "--project-pipeline-interfaces",
                dest=PROJECT_PL_ARG,
                metavar="YAML",
                default=None,
                nargs="+",
                type=str,
                help="Path to looper project config file",
            )
            # help="Path to the output directory"
            subparser.add_argument(
                "-o",
                "--output-dir",
                dest="output_dir",
                metavar="DIR",
                default=None,
                type=str,
                help=argparse.SUPPRESS,
            )
            # "Submission subdirectory name"
            subparser.add_argument(
                "--submission-subdir", metavar="DIR", help=argparse.SUPPRESS
            )
            # "Results subdirectory name"
            subparser.add_argument(
                "--results-subdir", metavar="DIR", help=argparse.SUPPRESS
            )
            # "Sample attribute for pipeline interface sources"
            subparser.add_argument(
                "--pipeline-interfaces-key", metavar="K", help=argparse.SUPPRESS
            )
            # "Paths to pipeline interface files"
            subparser.add_argument(
                "--pipeline-interfaces",
                metavar="P",
                nargs="+",
                action="append",
                help=argparse.SUPPRESS,
            )

        for subparser in [
            run_subparser,
            rerun_subparser,
            table_subparser,
            report_subparser,
            destroy_subparser,
            check_subparser,
            clean_subparser,
            collate_subparser,
            inspect_subparser,
            link_subparser,
        ]:
            fetch_samples_group = subparser.add_argument_group(
                "sample selection arguments",
                "Specify samples to include or exclude based on sample attribute values",
            )
            fetch_samples_group.add_argument(
                "-l",
                "--limit",
                default=None,
                metavar="N",
                type=html_range(min_val=1, max_val="num_samples", value="num_samples"),
                help="Limit to n samples",
            )
            fetch_samples_group.add_argument(
                "-k",
                "--skip",
                default=None,
                metavar="N",
                type=html_range(min_val=1, max_val="num_samples", value="num_samples"),
                help="Skip samples by numerical index",
            )

            fetch_samples_group.add_argument(
                f"--{SAMPLE_SELECTION_ATTRIBUTE_OPTNAME}",
                default="toggle",
                metavar="ATTR",
                help="Attribute for sample exclusion OR inclusion",
            )

            protocols = fetch_samples_group.add_mutually_exclusive_group()
            protocols.add_argument(
                f"--{SAMPLE_EXCLUSION_OPTNAME}",
                nargs="*",
                metavar="E",
                help="Exclude samples with these values",
            )
            protocols.add_argument(
                f"--{SAMPLE_INCLUSION_OPTNAME}",
                nargs="*",
                metavar="I",
                help="Include only samples with these values",
            )
            fetch_samples_group.add_argument(
                f"--{SAMPLE_SELECTION_FLAG_OPTNAME}",
                default=None,
                nargs="*",
                metavar="SELFLAG",
                help="Include samples with this flag status, e.g. completed",
            )

            fetch_samples_group.add_argument(
                f"--{SAMPLE_EXCLUSION_FLAG_OPTNAME}",
                default=None,
                nargs="*",
                metavar="EXCFLAG",
                help="Exclude samples with this flag status, e.g. completed",
            )

            subparser.add_argument(
                "-a",
                "--amend",
                nargs="+",
                metavar="A",
                help="List of amendments to activate",
            )
        for subparser in [
            report_subparser,
            table_subparser,
            check_subparser,
            destroy_subparser,
            link_subparser,
        ]:
            subparser.add_argument(
                "--project",
                help="Process project-level pipelines",
                action="store_true",
                default=False,
            )
        inspect_subparser.add_argument(
            "--sample-names",
            help="Names of the samples to inspect",
            nargs="*",
            default=None,
        )

        inspect_subparser.add_argument(
            "--attr-limit",
            help="Number of attributes to display",
            type=int,
        )
        result.append(parser)
    return result


def opt_attr_pair(name: str) -> Tuple[str, str]:
    return f"--{name}", name.replace("-", "_")


def validate_post_parse(args: argparse.Namespace) -> List[str]:
    problems = []
    used_exclusives = [
        opt
        for opt, attr in map(
            opt_attr_pair,
            [
                "skip",
                "limit",
                SAMPLE_EXCLUSION_OPTNAME,
                SAMPLE_INCLUSION_OPTNAME,
            ],
        )
        if getattr(args, attr, None)
    ]
    if len(used_exclusives) > 1:
        problems.append(
            f"Used multiple mutually exclusive options: {', '.join(used_exclusives)}"
        )
    return problems


def _proc_resources_spec(args):
    """
    Process CLI-sources compute setting specification. There are two sources
    of compute settings in the CLI alone:
        * YAML file (--settings argument)
        * itemized compute settings (--compute argument)

    The itemized compute specification is given priority

    :param argparse.Namespace: arguments namespace
    :return Mapping[str, str]: binding between resource setting name and value
    :raise ValueError: if interpretation of the given specification as encoding
        of key-value pairs fails
    """
    spec = getattr(args, "compute", None)
    try:
        settings_data = read_yaml_file(args.settings) or {}
    except yaml.YAMLError:
        _LOGGER.warning(
            "Settings file ({}) does not follow YAML format,"
            " disregarding".format(args.settings)
        )
        settings_data = {}
    if not spec:
        return settings_data
    pairs = [(kv, kv.split("=")) for kv in spec]
    bads = []
    for orig, pair in pairs:
        try:
            k, v = pair
        except ValueError:
            bads.append(orig)
        else:
            settings_data[k] = v
    if bads:
        raise ValueError(
            "Could not correctly parse itemized compute specification. "
            "Correct format: " + EXAMPLE_COMPUTE_SPEC_FMT
        )
    return settings_data


def main(test_args=None):
    """Primary workflow"""
    global _LOGGER

    parser, aux_parser = build_parser()
    aux_parser.suppress_defaults()

    if test_args:
        args, remaining_args = parser.parse_known_args(args=test_args)
    else:
        args, remaining_args = parser.parse_known_args()

    cli_use_errors = validate_post_parse(args)
    if cli_use_errors:
        parser.print_help(sys.stderr)
        parser.error(
            f"{len(cli_use_errors)} CLI use problem(s): {', '.join(cli_use_errors)}"
        )
    if args.command is None:
        parser.print_help(sys.stderr)
        sys.exit(1)

    if args.command == "init":
        return int(
            not initiate_looper_config(
                dotfile_path(),
                args.pep_config,
                args.output_dir,
                args.sample_pipeline_interfaces,
                args.project_pipeline_interfaces,
                args.force,
            )
        )

    if args.command == "init-piface":
        sys.exit(int(not init_generic_pipeline()))

    _LOGGER = logmuse.logger_via_cli(args, make_root=True)
    _LOGGER.info("Looper version: {}\nCommand: {}".format(__version__, args.command))

    if "config_file" in vars(args):
        if args.config_file is None:
            looper_cfg_path = os.path.relpath(dotfile_path(), start=os.curdir)
            try:
                if args.looper_config:
                    looper_config_dict = read_looper_config_file(args.looper_config)
                else:
                    looper_config_dict = read_looper_dotfile()
                    _LOGGER.info(f"Using looper config ({looper_cfg_path}).")

                for looper_config_key, looper_config_item in looper_config_dict.items():
                    setattr(args, looper_config_key, looper_config_item)

            except OSError:
                parser.print_help(sys.stderr)
                _LOGGER.warning(
                    f"Looper config file does not exist. Use looper init to create one at {looper_cfg_path}."
                )
                sys.exit(1)
        else:
            _LOGGER.warning(
                "This PEP configures looper through the project config. This approach is deprecated and will "
                "be removed in future versions. Please use a looper config file. For more information see "
                "looper.databio.org/en/latest/looper-config"
            )

    args = enrich_args_via_cfg(args, aux_parser, test_args)

    # If project pipeline interface defined in the cli, change name to: "pipeline_interface"
    if vars(args)[PROJECT_PL_ARG]:
        args.pipeline_interfaces = vars(args)[PROJECT_PL_ARG]

    if len(remaining_args) > 0:
        _LOGGER.warning(
            "Unrecognized arguments: {}".format(
                " ".join([str(x) for x in remaining_args])
            )
        )

    divcfg = (
        select_divvy_config(filepath=args.divvy) if hasattr(args, "divvy") else None
    )

    # Ignore flags if user is selecting or excluding on flags:
    if args.sel_flag or args.exc_flag:
        args.ignore_flags = True

    # Initialize project
    if is_registry_path(args.config_file):
        if vars(args)[SAMPLE_PL_ARG]:
            p = Project(
                amendments=args.amend,
                divcfg_path=divcfg,
                runp=args.command == "runp",
                project_dict=PEPHubClient()._load_raw_pep(
                    registry_path=args.config_file
                ),
                **{
                    attr: getattr(args, attr) for attr in CLI_PROJ_ATTRS if attr in args
                },
            )
        else:
            raise MisconfigurationException(
                f"`sample_pipeline_interface` is missing. Provide it in the parameters."
            )
    else:
        try:
            p = Project(
                cfg=args.config_file,
                amendments=args.amend,
                divcfg_path=divcfg,
                runp=args.command == "runp",
                **{
                    attr: getattr(args, attr) for attr in CLI_PROJ_ATTRS if attr in args
                },
            )
        except yaml.parser.ParserError as e:
            _LOGGER.error(f"Project config parse failed -- {e}")
            sys.exit(1)

    selected_compute_pkg = p.selected_compute_package or DEFAULT_COMPUTE_RESOURCES_NAME
    if p.dcc is not None and not p.dcc.activate_package(selected_compute_pkg):
        _LOGGER.info(
            "Failed to activate '{}' computing package. "
            "Using the default one".format(selected_compute_pkg)
        )

    with ProjectContext(
        prj=p,
        selector_attribute=args.sel_attr,
        selector_include=args.sel_incl,
        selector_exclude=args.sel_excl,
        selector_flag=args.sel_flag,
        exclusion_flag=args.exc_flag,
    ) as prj:
        if args.command in ["run", "rerun"]:
            run = Runner(prj)
            try:
                compute_kwargs = _proc_resources_spec(args)
                return run(args, rerun=(args.command == "rerun"), **compute_kwargs)
            except SampleFailedException:
                sys.exit(1)
            except IOError:
                _LOGGER.error(
                    "{} pipeline_interfaces: '{}'".format(
                        prj.__class__.__name__, prj.pipeline_interface_sources
                    )
                )
                raise

        if args.command == "runp":
            compute_kwargs = _proc_resources_spec(args)
            collate = Collator(prj)
            collate(args, **compute_kwargs)
            return collate.debug

        if args.command == "destroy":
            return Destroyer(prj)(args)

        # pipestat support introduces breaking changes and pipelines run
        # with no pipestat reporting would not be compatible with
        # commands: table, report and check. Therefore we plan maintain
        # the old implementations for a couple of releases.
        # if hasattr(args, "project"):
        #     use_pipestat = (
        #         prj.pipestat_configured_project
        #         if args.project
        #         else prj.pipestat_configured
        #     )
        use_pipestat = (
            prj.pipestat_configured_project if args.project else prj.pipestat_configured
        )
        if args.command == "table":
            if use_pipestat:
                Tabulator(prj)(args)
            else:
                raise PipestatConfigurationException("table")

        if args.command == "report":
            if use_pipestat:
                Reporter(prj)(args)
            else:
                raise PipestatConfigurationException("report")

        if args.command == "link":
            if use_pipestat:
                Linker(prj)(args)
            else:
                raise PipestatConfigurationException("link")

        if args.command == "check":
            if use_pipestat:
                return Checker(prj)(args)
            else:
                raise PipestatConfigurationException("check")

        if args.command == "clean":
            return Cleaner(prj)(args)

        if args.command == "inspect":
            inspect_project(p, args.sample_names, args.attr_limit)
            from warnings import warn

            warn(
                "The inspect feature has moved to eido and will be removed in the future release of looper. "
                "Use `eido inspect` from now on.",
            )



import logmuse
import os
import sys
import yaml
from yaml import SafeLoader
from ubiquerg import is_writable, VersionInHelpParser
from .const import (
    DEFAULT_COMPUTE_RESOURCES_NAME,
    DEFAULT_CONFIG_FILEPATH,
)
from .divvy import select_divvy_config, ComputingConfiguration, divvy_init


def build_argparser():
    """
    Builds argument parser.

    :return argparse.ArgumentParser
    """

    banner = (
        "%(prog)s - write compute job scripts that can be submitted to "
        "any computing resource"
    )
    additional_description = "\nhttps://divvy.databio.org"

    parser = VersionInHelpParser(
        prog="divvy",
        description=banner,
        epilog=additional_description,
        # version=__version__,
    )

    subparsers = parser.add_subparsers(dest="command")

    def add_subparser(cmd, description):
        return subparsers.add_parser(cmd, description=description, help=description)

    subparser_messages = {
        "init": "Initialize a new divvy config file",
        "list": "List available compute packages",
        "write": "Write a job script",
        "submit": "Write and then submit a job script",
        "inspect": "Inspect compute package",
    }

    sps = {}
    for cmd, desc in subparser_messages.items():
        sps[cmd] = add_subparser(cmd, desc)
        # sps[cmd].add_argument(
        #     "config", nargs="?", default=None,
        #     help="Divvy configuration file.")

    for sp in [sps["list"], sps["write"], sps["submit"], sps["inspect"]]:
        sp.add_argument(
            "config", nargs="?", default=None, help="Divvy configuration file."
        )

    sps["init"].add_argument("config", default=None, help="Divvy configuration file.")

    for sp in [sps["inspect"]]:
        sp.add_argument(
            "-p",
            "--package",
            default=DEFAULT_COMPUTE_RESOURCES_NAME,
            help="Select from available compute packages",
        )

    for sp in [sps["write"], sps["submit"]]:
        sp.add_argument(
            "-s",
            "--settings",
            help="YAML file with job settings to populate the template",
        )

        sp.add_argument(
            "-p",
            "--package",
            default=DEFAULT_COMPUTE_RESOURCES_NAME,
            help="Select from available compute packages",
        )

        sp.add_argument(
            "-c",
            "--compute",
            nargs="+",
            default=None,
            help="Extra key=value variable pairs",
        )

        # sp.add_argument(
        #         "-t", "--template",
        #         help="Provide a template file (not yet implemented).")

        sp.add_argument(
            "-o", "--outfile", required=False, default=None, help="Output filepath"
        )

    return parser


def main():
    """Primary workflow for divvy CLI"""

    parser = logmuse.add_logging_options(build_argparser())
    # args, remaining_args = parser.parse_known_args()
    args = parser.parse_args()

    logger_kwargs = {"level": args.verbosity, "devmode": args.logdev}
    logmuse.init_logger("yacman", **logger_kwargs)
    global _LOGGER
    _LOGGER = logmuse.logger_via_cli(args)

    if not args.command:
        parser.print_help()
        _LOGGER.error("No command given")
        sys.exit(1)

    if args.command == "init":
        divcfg = args.config
        _LOGGER.debug("Initializing divvy configuration")
        is_writable(os.path.dirname(divcfg), check_exist=False)
        divvy_init(divcfg, DEFAULT_CONFIG_FILEPATH)
        sys.exit(0)

    _LOGGER.debug("Divvy config: {}".format(args.config))
    divcfg = select_divvy_config(args.config)
    _LOGGER.info("Using divvy config: {}".format(divcfg))
    dcc = ComputingConfiguration(filepath=divcfg)

    if args.command == "list":
        # Output header via logger and content via print so the user can
        # redirect the list from stdout if desired without the header as clutter
        _LOGGER.info("Available compute packages:\n")
        print("{}".format("\n".join(dcc.list_compute_packages())))
        sys.exit(1)

    if args.command == "inspect":
        # Output contents of selected compute package
        _LOGGER.info("Your compute package template for: " + args.package + "\n")
        found = False
        for pkg_name, pkg in dcc.compute_packages.items():
            if pkg_name == args.package:
                found = True
                with open(pkg.submission_template, "r") as f:
                    print(f.read())
                _LOGGER.info("Submission command is: " + pkg.submission_command + "\n")
                if pkg_name == "docker":
                    print("Docker args are: " + pkg.docker_args)

        if not found:
            _LOGGER.info("Package not found. Use 'divvy list' to see list of packages.")
        sys.exit(1)

    # Any non-divvy arguments will be passed along as key-value pairs
    # that can be used to populate the template.
    # keys = [str.replace(x, "--", "") for x in remaining_args[::2]]
    # cli_vars = dict(zip(keys, remaining_args[1::2]))
    if args.compute:
        cli_vars = {y[0]: y[1] for y in [x.split("=") for x in args.compute]}
    else:
        cli_vars = {}

    if args.command == "write" or args.command == "submit":
        try:
            dcc.activate_package(args.package)
        except AttributeError:
            parser.print_help(sys.stderr)
            sys.exit(1)

        if args.settings:
            _LOGGER.info("Loading settings file: %s", args.settings)
            with open(args.settings, "r") as f:
                vars_groups = [cli_vars, yaml.load(f, SafeLoader)]
        else:
            vars_groups = [cli_vars]

        _LOGGER.debug(vars_groups)
        if args.command == "write":
            dcc.write_script(args.outfile, vars_groups)
        elif args.command == "submit":
            dcc.submit(args.outfile, vars_groups)



""" Helpers without an obvious logical home. """

import argparse
from collections import defaultdict, namedtuple
import glob
import itertools
from logging import getLogger
import os
import sys
from typing import *
import re

import jinja2
import yaml
from peppy import Project as peppyProject
from peppy.const import *
from ubiquerg import convert_value, expandpath, parse_registry_path
from pephubclient.constants import RegistryPath
from pydantic.error_wrappers import ValidationError

from .const import *
from .exceptions import MisconfigurationException, RegistryPathException

_LOGGER = getLogger(__name__)


def fetch_flag_files(prj=None, results_folder="", flags=FLAGS):
    """
    Find all flag file paths for the given project.

    :param Project | AttributeDict prj: full Project or AttributeDict with
        similar metadata and access/usage pattern
    :param str results_folder: path to results folder, corresponding to the
        1:1 sample:folder notion that a looper Project has. That is, this
        function uses the assumption that if results_folder rather than project
        is provided, the structure of the file tree rooted at results_folder is
        such that any flag files to be found are not directly within rootdir but
        are directly within on of its first layer of subfolders.
    :param Iterable[str] | str flags: Collection of flag names or single flag
        name for which to fetch files
    :return Mapping[str, list[str]]: collection of filepaths associated with
        particular flag for samples within the given project
    :raise TypeError: if neither or both of project and rootdir are given
    """

    if not (prj or results_folder) or (prj and results_folder):
        raise TypeError("Need EITHER project OR rootdir")

    # Just create the filenames once, and pair once with flag name.
    flags = [flags] if isinstance(flags, str) else list(flags)
    flagfile_suffices = ["*{}.flag".format(f) for f in flags]
    flag_suffix_pairs = list(zip(flags, flagfile_suffices))

    # Collect the flag file paths by flag name.
    files_by_flag = defaultdict(list)

    if prj is None:
        for flag, suffix in flag_suffix_pairs:
            flag_expr = os.path.join(results_folder, "*", suffix)
            flags_present = glob.glob(flag_expr)
            files_by_flag[flag] = flags_present
    else:
        # Iterate over samples to collect flag files.
        for s in prj.samples:
            folder = sample_folder(prj, s)
            # Check each candidate flag for existence, collecting if present.
            for flag, suffix in flag_suffix_pairs:
                flag_expr = os.path.join(folder, suffix)
                flags_present = glob.glob(flag_expr)
                files_by_flag[flag].extend(flags_present)

    return files_by_flag


def fetch_sample_flags(prj, sample, pl_name, flag_dir=None):
    """
    Find any flag files present for a sample associated with a project

    :param looper.Project prj: project of interest
    :param peppy.Sample sample: sample object of interest
    :param str pl_name: name of the pipeline for which flag(s) should be found
    :return Iterable[str]: collection of flag file path(s) associated with the
        given sample for the given project
    """
    sfolder = flag_dir or sample_folder(prj=prj, sample=sample)
    if not os.path.isdir(sfolder):
        _LOGGER.debug(
            "Results folder ({}) doesn't exist for sample {}".format(
                sfolder, str(sample)
            )
        )
        return []
    folder_contents = [os.path.join(sfolder, f) for f in os.listdir(sfolder)]
    return [
        x
        for x in folder_contents
        if os.path.splitext(x)[1] == ".flag" and os.path.basename(x).startswith(pl_name)
    ]


def get_sample_status(sample, flags):
    """
    get a sample status

    """

    statuses = []

    for f in flags:
        basename = os.path.basename(f)
        status = os.path.splitext(basename)[0].split("_")[-1]
        if sample in basename:
            statuses.append(status.upper())

    if len(statuses) > 1:
        _LOGGER.warning(f"Multiple status flags found for {sample}")

    if statuses == []:
        return None

    return statuses[0]


def grab_project_data(prj):
    """
    From the given Project, grab Sample-independent data.

    There are some aspects of a Project of which it's beneficial for a Sample
    to be aware, particularly for post-hoc analysis. Since Sample objects
    within a Project are mutually independent, though, each doesn't need to
    know about any of the others. A Project manages its, Sample instances,
    so for each Sample knowledge of Project data is limited. This method
    facilitates adoption of that conceptual model.

    :param Project prj: Project from which to grab data
    :return Mapping: Sample-independent data sections from given Project
    """
    if not prj:
        return {}

    try:
        return prj[CONFIG_KEY]
    except KeyError:
        _LOGGER.debug("Project lacks section '%s', skipping", CONFIG_KEY)


def sample_folder(prj, sample):
    """
    Get the path to this Project's root folder for the given Sample.

    :param AttributeDict | Project prj: project with which sample is associated
    :param Mapping sample: Sample or sample data for which to get root output
        folder path.
    :return str: this Project's root folder for the given Sample
    """
    return os.path.join(prj.results_folder, sample[prj.sample_table_index])


def get_file_for_project(prj, pipeline_name, appendix=None, directory=None):
    """
    Create a path to the file for the current project.
    Takes the possibility of amendment being activated at the time

    Format of the output path:
    {output_dir}/{directory}/{p.name}_{pipeline_name}_{active_amendments}_{appendix}

    :param looper.Project prj: project object
    :param str pipeline_name: name of the pipeline to get the file for
    :param str appendix: the appendix of the file to create the path for,
        like 'objs_summary.tsv' for objects summary file
    :return str: path to the file
    """
    fp = os.path.join(
        prj.output_dir, directory or "", f"{prj[NAME_KEY]}_{pipeline_name}"
    )
    if hasattr(prj, "amendments") and getattr(prj, "amendments"):
        fp += f"_{'_'.join(prj.amendments)}"
    fp += f"_{appendix}"
    return fp


def get_file_for_project_old(prj, appendix):
    """
    Create a path to the file for the current project.
    Takes the possibility of amendment being activated at the time

    :param looper.Project prj: project object
    :param str appendix: the appendix of the file to create the path for,
        like 'objs_summary.tsv' for objects summary file
    :return str: path to the file
    """
    fp = os.path.join(prj.output_dir, prj[NAME_KEY])
    if hasattr(prj, AMENDMENTS_KEY) and getattr(prj, AMENDMENTS_KEY):
        fp += "_" + "_".join(getattr(prj, AMENDMENTS_KEY))
    fp += "_" + appendix
    return fp


def jinja_render_template_strictly(template, namespaces):
    """
    Render a command string in the provided namespaces context.

    Strictly, which means that all the requested attributes must be
    available in the namespaces

    :param str template: command template do be filled in with the
        variables in the provided namespaces. For example:
        "prog.py --name {project.name} --len {sample.len}"
    :param Mapping[Mapping[str] namespaces: context for command rendering.
        Possible namespaces are: looper, project, sample, pipeline
    :return str: rendered command
    """

    def _finfun(x):
        """
        A callable that can be used to process the result of a variable
        expression before it is output. Joins list elements
        """
        return " ".join(x) if isinstance(x, list) else x

    env = jinja2.Environment(
        undefined=jinja2.StrictUndefined,
        variable_start_string="{",
        variable_end_string="}",
        finalize=_finfun,
    )
    templ_obj = env.from_string(template)
    try:
        rendered = templ_obj.render(**namespaces)
    except jinja2.exceptions.UndefinedError as e:
        _LOGGER.error("Error populating command template: " + str(e))
        _LOGGER.debug(f"({', '.join(list(namespaces.keys()))}) missing for ")
        _LOGGER.debug(f"Template: '{template}'")
        raise e
    _LOGGER.debug("rendered arg str: {}".format(rendered))
    return rendered


def read_yaml_file(filepath):
    """
    Read a YAML file

    :param str filepath: path to the file to read
    :return dict: read data
    """
    data = None
    if os.path.exists(filepath):
        with open(filepath, "r") as f:
            data = yaml.safe_load(f)
    return data


def enrich_args_via_cfg(parser_args, aux_parser, test_args=None):
    """
    Read in a looper dotfile and set arguments.

    Priority order: CLI > dotfile/config > parser default

    :param argparse.Namespace parser_args: parsed args by the original parser
    :param argparse.Namespace aux_parser: parsed args by the a parser
        with defaults suppressed
    :return argparse.Namespace: selected argument values
    """
    cfg_args_all = (
        _get_subcommand_args(parser_args)
        if os.path.exists(parser_args.config_file)
        else dict()
    )
    result = argparse.Namespace()
    if test_args:
        cli_args, _ = aux_parser.parse_known_args(args=test_args)

    else:
        cli_args, _ = aux_parser.parse_known_args()

    for dest in vars(parser_args):
        if dest not in POSITIONAL or not hasattr(result, dest):
            if dest in cli_args:
                x = getattr(cli_args, dest)
                r = convert_value(x) if isinstance(x, str) else x
            elif cfg_args_all is not None and dest in cfg_args_all:
                if isinstance(cfg_args_all[dest], list):
                    r = [convert_value(i) for i in cfg_args_all[dest]]
                else:
                    r = convert_value(cfg_args_all[dest])
            else:
                r = getattr(parser_args, dest)
            setattr(result, dest, r)
    return result


def _get_subcommand_args(parser_args):
    """
    Get the union of values for the subcommand arguments from
    Project.looper, Project.looper.cli.<subcommand> and Project.looper.cli.all.
    If any are duplicated, the above is the selection priority order.

    Additionally, convert the options strings to destinations (replace '-'
    with '_'), which strongly relies on argument parser using default
    destinations.

    :param argparser.Namespace parser_args: argument namespace
    :return dict: mapping of argument destinations to their values
    """
    args = dict()
    cfg = peppyProject(
        parser_args.config_file,
        defer_samples_creation=True,
        amendments=parser_args.amend,
    )
    if (
        CONFIG_KEY in cfg
        and LOOPER_KEY in cfg[CONFIG_KEY]
        and CLI_KEY in cfg[CONFIG_KEY][LOOPER_KEY]
    ):
        try:
            cfg_args = cfg[CONFIG_KEY][LOOPER_KEY][CLI_KEY] or dict()
            args = (
                cfg_args[ALL_SUBCMD_KEY] or dict()
                if ALL_SUBCMD_KEY in cfg_args
                else dict()
            )
            args.update(
                cfg_args[parser_args.command] or dict()
                if parser_args.command in cfg_args
                else dict()
            )
        except (TypeError, KeyError, AttributeError, ValueError) as e:
            raise MisconfigurationException(
                "Invalid '{}.{}' section in the config. Caught exception: {}".format(
                    LOOPER_KEY, CLI_KEY, getattr(e, "message", repr(e))
                )
            )
    if CONFIG_KEY in cfg and LOOPER_KEY in cfg[CONFIG_KEY]:
        try:
            if CLI_KEY in cfg[CONFIG_KEY][LOOPER_KEY]:
                del cfg[CONFIG_KEY][LOOPER_KEY][CLI_KEY]
            args.update(cfg[CONFIG_KEY][LOOPER_KEY])
        except (TypeError, KeyError, AttributeError, ValueError) as e:
            raise MisconfigurationException(
                "Invalid '{}' section in the config. Caught exception: {}".format(
                    LOOPER_KEY, getattr(e, "message", repr(e))
                )
            )
    args = {k.replace("-", "_"): v for k, v in args.items()} if args else None
    return args


def init_generic_pipeline():
    """
    Create generic pipeline interface
    """
    try:
        os.makedirs("pipeline")
    except FileExistsError:
        pass

    # Destination one level down from CWD in pipeline folder
    dest_file = os.path.join(os.getcwd(), "pipeline", LOOPER_GENERIC_PIPELINE)

    # Create Generic Pipeline Interface
    generic_pipeline_dict = {
        "pipeline_name": "default_pipeline_name",
        "pipeline_type": "sample",
        "output_schema": "output_schema.yaml",
        "var_templates": {"pipeline": "{looper.piface_dir}/pipeline.sh"},
        "command_template": "{pipeline.var_templates.pipeline} {sample.file} "
        "--output-parent {looper.sample_output_folder}",
    }

    # Write file
    if not os.path.exists(dest_file):
        with open(dest_file, "w") as file:
            yaml.dump(generic_pipeline_dict, file)
        print(f"Pipeline interface successfully created at: {dest_file}")
    else:
        print(
            f"Pipeline interface file already exists `{dest_file}`. Skipping creation.."
        )

    # Create Generic Output Schema
    dest_file = os.path.join(os.getcwd(), "pipeline", LOOPER_GENERIC_OUTPUT_SCHEMA)
    generic_output_schema_dict = {
        "pipeline_name": "default_pipeline_name",
        "samples": {
            "number_of_lines": {
                "type": "integer",
                "description": "Number of lines in the input file.",
            }
        },
    }
    # Write file
    if not os.path.exists(dest_file):
        with open(dest_file, "w") as file:
            yaml.dump(generic_output_schema_dict, file)
        print(f"Output schema successfully created at: {dest_file}")
    else:
        print(f"Output schema file already exists `{dest_file}`. Skipping creation..")

    # Create Generic countlines.sh
    dest_file = os.path.join(os.getcwd(), "pipeline", LOOPER_GENERIC_COUNT_LINES)
    shell_code = """#!/bin/bash
linecount=`wc -l $1 | sed -E 's/^[[:space:]]+//' | cut -f1 -d' '`
pipestat report -r $2 -i 'number_of_lines' -v $linecount -c $3
echo "Number of lines: $linecount"
    """
    if not os.path.exists(dest_file):
        with open(dest_file, "w") as file:
            file.write(shell_code)
        print(f"count_lines.sh successfully created at: {dest_file}")
    else:
        print(f"count_lines.sh file already exists `{dest_file}`. Skipping creation..")

    return True


def read_looper_dotfile():
    """
    Read looper config file
    :return str: path to the config file read from the dotfile
    :raise MisconfigurationException: if the dotfile does not consist of the
        required key pointing to the PEP
    """
    dot_file_path = dotfile_path(must_exist=True)
    return read_looper_config_file(looper_config_path=dot_file_path)


def initiate_looper_config(
    looper_config_path: str,
    pep_path: str = None,
    output_dir: str = None,
    sample_pipeline_interfaces: Union[List[str], str] = None,
    project_pipeline_interfaces: Union[List[str], str] = None,
    force=False,
):
    """
    Initialize looper config file

    :param str looper_config_path: absolute path to the file to initialize
    :param str pep_path: path to the PEP to be used in pipeline
    :param str output_dir: path to the output directory
    :param str|list sample_pipeline_interfaces: path or list of paths to sample pipeline interfaces
    :param str|list project_pipeline_interfaces: path or list of paths to project pipeline interfaces
    :param bool force: whether the existing file should be overwritten
    :return bool: whether the file was initialized
    """
    if os.path.exists(looper_config_path) and not force:
        print(f"Can't initialize, file exists: {looper_config_path}")
        return False

    if pep_path:
        if is_registry_path(pep_path):
            pass
        else:
            pep_path = expandpath(pep_path)
            if not os.path.isabs(pep_path):
                pep_path = os.path.join(os.path.dirname(looper_config_path), pep_path)
            assert os.path.exists(pep_path), OSError(
                "Provided config path is invalid. You must provide path "
                f"that is either absolute or relative to: {os.path.dirname(looper_config_path)}"
            )
    else:
        pep_path = "example/pep/path"

    if not output_dir:
        output_dir = "."

    looper_config_dict = {
        "pep_config": os.path.relpath(pep_path),
        "output_dir": output_dir,
        "pipeline_interfaces": {
            "sample": sample_pipeline_interfaces,
            "project": project_pipeline_interfaces,
        },
    }

    with open(looper_config_path, "w") as dotfile:
        yaml.dump(looper_config_dict, dotfile)
    print(f"Initialized looper config file: {looper_config_path}")
    return True


def read_looper_config_file(looper_config_path: str) -> dict:
    """
    Read Looper config file which includes:
    - PEP config (local path or pephub registry path)
    - looper output dir
    - looper pipeline interfaces

    :param str looper_config_path: path to looper config path
    :return dict: looper config file content
    :raise MisconfigurationException: incorrect configuration.
    """
    return_dict = {}
    with open(looper_config_path, "r") as dotfile:
        dp_data = yaml.safe_load(dotfile)

    if PEP_CONFIG_KEY in dp_data:
        # Looper expects the config path to live at looper.config_file
        # However, user may wish to access the pep at looper.pep_config
        return_dict[PEP_CONFIG_FILE_KEY] = dp_data[PEP_CONFIG_KEY]
        return_dict[PEP_CONFIG_KEY] = dp_data[PEP_CONFIG_KEY]

    # TODO: delete it in looper 2.0
    elif DOTFILE_CFG_PTH_KEY in dp_data:
        return_dict[PEP_CONFIG_FILE_KEY] = dp_data[DOTFILE_CFG_PTH_KEY]

    else:
        raise MisconfigurationException(
            f"Looper dotfile ({looper_config_path}) is missing '{PEP_CONFIG_KEY}' key"
        )

    if OUTDIR_KEY in dp_data:
        return_dict[OUTDIR_KEY] = dp_data[OUTDIR_KEY]
    else:
        _LOGGER.warning(
            f"{OUTDIR_KEY} is not defined in looper config file ({looper_config_path})"
        )

    if PIPESTAT_KEY in dp_data:
        return_dict[PIPESTAT_KEY] = dp_data[PIPESTAT_KEY]

    if PIPELINE_INTERFACES_KEY in dp_data:
        dp_data.setdefault(PIPELINE_INTERFACES_KEY, {})
        return_dict[SAMPLE_PL_ARG] = dp_data.get(PIPELINE_INTERFACES_KEY).get("sample")
        return_dict[PROJECT_PL_ARG] = dp_data.get(PIPELINE_INTERFACES_KEY).get(
            "project"
        )

    else:
        _LOGGER.warning(
            f"{PIPELINE_INTERFACES_KEY} is not defined in looper config file ({looper_config_path})"
        )
        dp_data.setdefault(PIPELINE_INTERFACES_KEY, {})

    config_dir_path = os.path.dirname(os.path.abspath(looper_config_path))

    # Expand paths in case ENV variables are used
    for k, v in return_dict.items():
        if isinstance(v, str):
            v = expandpath(v)
            if not os.path.isabs(v) and not is_registry_path(v):
                return_dict[k] = os.path.join(config_dir_path, v)
            else:
                return_dict[k] = v

    return return_dict


def dotfile_path(directory=os.getcwd(), must_exist=False):
    """
    Get the path to the looper dotfile

    If file existence is forced this function will look for it in
    the directory parents

    :param str directory: directory path to start the search in
    :param bool must_exist: whether the file must exist
    :return str: path to the dotfile
    :raise OSError: if the file does not exist
    """
    cur_dir = directory
    if not must_exist:
        return os.path.join(cur_dir, LOOPER_DOTFILE_NAME)
    while True:
        parent_dir = os.path.dirname(cur_dir)
        if LOOPER_DOTFILE_NAME in os.listdir(cur_dir):
            return os.path.join(cur_dir, LOOPER_DOTFILE_NAME)
        if cur_dir == parent_dir:
            # root, file does not exist
            raise OSError(
                "Looper dotfile ({}) not found in '{}' and all "
                "its parents".format(LOOPER_DOTFILE_NAME, directory)
            )
        cur_dir = parent_dir


def is_registry_path(input_string: str) -> bool:
    """
    Check if input is a registry path to pephub
    :param str input_string: path to the PEP (or registry path)
    :return bool: True if input is a registry path
    """
    try:
        if input_string.endswith(".yaml"):
            return False
    except AttributeError:
        raise RegistryPathException(
            msg=f"Malformed registry path. Unable to parse {input_string} as a registry path."
        )
    try:
        registry_path = RegistryPath(**parse_registry_path(input_string))
    except (ValidationError, TypeError):
        return False
    return True


class NatIntervalException(Exception):
    """Subtype for errors specifically related to natural number interval"""

    pass


class NatIntervalInclusive(object):
    def __init__(self, lo: int, hi: int):
        super().__init__()
        self._lo = lo
        self._hi = hi
        problems = self._invalidations()
        if problems:
            raise NatIntervalException(
                f"{len(problems)} issues with interval on natural numbers: {', '.join(problems)}"
            )

    def __eq__(self, other) -> bool:
        return type(other) == type(self) and self.to_tuple() == other.to_tuple()

    def __hash__(self) -> int:
        return hash(self.to_tuple())

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}: {self.to_tuple()}"

    def __str__(self) -> str:
        return f"{self.__class__.__name__}: {self.to_tuple()}"

    def to_tuple(self) -> Tuple[int, int]:
        return self.lo, self.hi

    @property
    def lo(self) -> int:
        return self._lo

    @property
    def hi(self) -> int:
        return self._hi

    def _invalidations(self) -> Iterable[str]:
        problems = []
        if self.lo < 1:
            problems.append(f"Interval must be on natural numbers: {self.lo}")
        if self.hi < self.lo:
            problems.append(
                f"Upper bound must not be less than lower bound: {self.hi} < {self.lo}"
            )
        return problems

    def to_range(self) -> Iterable[int]:
        return range(self.lo, self.hi + 1)

    @classmethod
    def from_string(cls, s: str, upper_bound: int) -> "IntRange":
        """
        Create an instance from a string, e.g. command-line argument.

        :param str s: The string to parse as an interval
        :param int upper_bound: the default upper bound
        """
        if upper_bound < 1:
            raise NatIntervalException(f"Upper bound must be positive: {upper_bound}")

        # Determine delimiter, invalidating presence of multiple occurrences.
        delim_histo = defaultdict(int)
        candidates = [":", "-"]
        for c in s:
            if c in candidates:
                delim_histo[c] += 1
        seps = [sep for sep, num_occ in delim_histo.items() if num_occ == 1]
        if len(seps) != 1:
            raise NatIntervalException(
                f"Did not find exactly one candidate delimiter with occurrence count of 1: {delim_histo}"
            )
        sep = seps[0]

        # Use the determined delimiter.
        lo, hi = s.split(sep)
        if lo == "" and hi == "":
            # We could do an interval like [1, upper_bound], but this is nonsensical as input.
            raise NatIntervalException(
                f"Parsed both lower and upper limit as empty from given arg: {s}"
            )
        try:
            lo = 1 if lo == "" else int(lo)
            hi = upper_bound if hi == "" else min(int(hi), upper_bound)
        except ValueError as e:
            raise NatIntervalException(str(e))
        return cls(lo, hi)


def desired_samples_range_limited(arg: str, num_samples: int) -> Iterable[int]:
    """
    Create a contiguous interval of natural numbers. Used for _positive_ selection of samples.

    Interpret given arg as upper bound (1-based) if it's a single value, but take the
    minimum of that and the given number of samples. If arg is parseable as a range,
    use that.

    :param str arg: CLI specification of a range of samples to use, or as the greatest
        1-based index of a sample to include
    :param int num_samples: what to use as the upper bound on the 1-based index interval
        if the given arg isn't a range but rather a single value.
    :return: an iterable of 1-based indices into samples to select
    """
    try:
        upper_bound = min(int(arg), num_samples)
    except ValueError:
        intv = NatIntervalInclusive.from_string(arg, upper_bound=num_samples)
    else:
        _LOGGER.debug("Limiting to {} of {} samples".format(upper_bound, num_samples))
        intv = NatIntervalInclusive(1, upper_bound)
    return intv.to_range()


def desired_samples_range_skipped(arg: str, num_samples: int) -> Iterable[int]:
    """
    Create a contiguous interval of natural numbers. Used for _negative_ selection of samples.

    :param str arg: CLI specification of a range of samples to use, or as the lowest
        1-based index of a sample to skip
    :param int num_samples: highest 1-based index of samples to include
    :return: an iterable of 1-based indices into samples to select
    """
    try:
        lower_bound = int(arg)
    except ValueError:
        intv = NatIntervalInclusive.from_string(arg, upper_bound=num_samples)
        lower = range(1, intv.lo)
        upper = range(intv.hi + 1, num_samples + 1)
        return itertools.chain(lower, upper)
    else:
        if num_samples <= lower_bound:
            return []
        intv = NatIntervalInclusive(lower_bound + 1, num_samples)
        return intv.to_range()


def write_submit_script(fp, content, data):
    """
    Write a submission script for divvy by populating a template with data.
    :param str fp: Path to the file to which to create/write submissions script.
    :param str content: Template for submission script, defining keys that
        will be filled by given data
    :param Mapping data: a "pool" from which values are available to replace
        keys in the template
    :return str: Path to the submission script
    """

    for k, v in data.items():
        placeholder = "{" + str(k).upper() + "}"
        content = content.replace(placeholder, str(v))

    keys_left = re.findall(r"!$\{(.+?)\}", content)
    if len(keys_left) > 0:
        _LOGGER.warning(
            "> Warning: %d submission template variables are not " "populated: '%s'",
            len(keys_left),
            str(keys_left),
        )

    if not fp:
        print(content)
        return content
    else:
        outdir = os.path.dirname(fp)
        if outdir and not os.path.isdir(outdir):
            os.makedirs(outdir)
        with open(fp, "w") as f:
            f.write(content)
        return fp



#! /usr/bin/env python

import os
import sys

from setuptools import setup

# Additional keyword arguments for setup().
extra = {}


# Ordinary dependencies
DEPENDENCIES = []
with open("requirements/requirements-all.txt", "r") as reqs_file:
    for line in reqs_file:
        if not line.strip():
            continue
        # DEPENDENCIES.append(line.split("=")[0].rstrip("<>"))
        DEPENDENCIES.append(line)


# numexpr for pandas
try:
    import numexpr
except ImportError:
    # No numexpr is OK for pandas.
    pass
else:
    # pandas 0.20.2 needs updated numexpr; the claim is 2.4.6, but that failed.
    DEPENDENCIES.append("numexpr>=2.6.2")
extra["install_requires"] = DEPENDENCIES


# Additional files to include with package
def get_static(name, condition=None):
    static = [
        os.path.join(name, f)
        for f in os.listdir(
            os.path.join(os.path.dirname(os.path.realpath(__file__)), name)
        )
    ]
    if condition is None:
        return static
    else:
        return [i for i in filter(lambda x: eval(condition), static)]


# scripts to be added to the $PATH
# scripts = get_static("scripts", condition="'.' in x")
# scripts removed (TO remove this)
scripts = None


with open("looper/_version.py", "r") as versionfile:
    version = versionfile.readline().split()[-1].strip("\"'\n")

with open("README.md") as f:
    long_description = f.read()

setup(
    name="looper",
    packages=["looper"],
    version=version,
    description="A pipeline submission engine that parses sample inputs and submits pipelines for each sample.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords="bioinformatics, sequencing, ngs",
    url="https://github.com/pepkit/looper",
    author="Nathan Sheffield, Vince Reuter, Michal Stolarczyk, Johanna Klughammer, Andre Rendeiro",
    license="BSD2",
    entry_points={
        "console_scripts": [
            "looper = looper.__main__:main",
            "divvy = looper.__main__:divvy_main",
        ],
    },
    scripts=scripts,
    package_data={"looper": ["submit_templates/*"]},
    include_package_data=True,
    test_suite="tests",
    tests_require=(["mock", "pytest"]),
    setup_requires=(
        ["pytest-runner"] if {"test", "pytest", "ptr"} & set(sys.argv) else []
    ),
    **extra
)



"""Project configuration, particularly for logging.

Project-scope constants may reside here, but more importantly, some setup here
will provide a logging infrastructure for all of the project's modules.
Individual modules and classes may provide separate configuration on a more
local level, but this will at least provide a foundation.

"""

import logmuse

logmuse.init_logger("looper")

from .divvy import ComputingConfiguration, select_divvy_config
from .divvy import DEFAULT_COMPUTE_RESOURCES_NAME
from .divvy import NEW_COMPUTE_KEY as COMPUTE_KEY

from ._version import __version__
from .conductor import (
    SubmissionConductor,
    write_submission_yaml,
)
from .plugins import (
    write_sample_yaml,
    write_sample_yaml_cwl,
    write_sample_yaml_prj,
    write_custom_template,
)
from .const import *
from .pipeline_interface import PipelineInterface
from .project import Project

# Not used here, but make this the main import interface between peppy and
# looper, so that other modules within this package need not worry about
# the locations of some of the peppy declarations. Effectively, concentrate
# the connection between peppy and looper here, to the extent possible.

__all__ = [
    "Project",
    "PipelineInterface",
    "SubmissionConductor",
    "ComputingConfiguration",
    "select_divvy_config",
]



""" Shared project constants """

import os

__author__ = "Databio lab"
__email__ = "nathan@code.databio.org"


__all__ = [
    "BUTTON_APPEARANCE_BY_FLAG",
    "TABLE_APPEARANCE_BY_FLAG",
    "VAR_TEMPL_KEY",
    "ID_COLNAME",
    "NO_DATA_PLACEHOLDER",
    "OUTKEY",
    "ALL_SUBCMD_KEY",
    "OUTDIR_KEY",
    "LOOPER_KEY",
    "COMPUTE_KEY",
    "PIPELINE_INTERFACES_KEY",
    "SIZE_DEP_VARS_KEY",
    "FLAGS",
    "DYN_VARS_KEY",
    "SAMPLE_YAML_PATH_KEY",
    "RESOURCES_KEY",
    "NOT_SUB_MSG",
    "EXTRA_KEY",
    "DEFAULT_CFG_PATH",
    "PIFACE_SCHEMA_SRC",
    "RESULTS_SUBDIR_KEY",
    "SUBMISSION_SUBDIR_KEY",
    "TEMPLATES_DIRNAME",
    "FILE_SIZE_COLNAME",
    "COMPUTE_PACKAGE_KEY",
    "INPUT_SCHEMA_KEY",
    "OUTPUT_SCHEMA_KEY",
    "EXAMPLE_COMPUTE_SPEC_FMT",
    "SAMPLE_PL_KEY",
    "PROJECT_PL_KEY",
    "CFG_ENV_VARS",
    "LOGGING_LEVEL",
    "PIFACE_KEY_SELECTOR",
    "SUBMISSION_FAILURE_MESSAGE",
    "IMAGE_EXTS",
    "PROFILE_COLNAMES",
    "SAMPLE_TOGGLE_ATTR",
    "LOOPER_DOTFILE_NAME",
    "POSITIONAL",
    "EXTRA_PROJECT_CMD_TEMPLATE",
    "EXTRA_SAMPLE_CMD_TEMPLATE",
    "SELECTED_COMPUTE_PKG",
    "CLI_PROJ_ATTRS",
    "DOTFILE_CFG_PTH_KEY",
    "DRY_RUN_KEY",
    "FILE_CHECKS_KEY",
    "CLI_KEY",
    "PRE_SUBMIT_HOOK_KEY",
    "PRE_SUBMIT_PY_FUN_KEY",
    "PRE_SUBMIT_CMD_KEY",
    "SUBMISSION_YAML_PATH_KEY",
    "SAMPLE_YAML_PRJ_PATH_KEY",
    "OBJECT_TYPES",
    "SAMPLE_CWL_YAML_PATH_KEY",
    "PIPESTAT_KEY",
    "NAMESPACE_ATTR_KEY",
    "DEFAULT_PIPESTAT_CONFIG_ATTR",
    "DEFAULT_PIPESTAT_RESULTS_FILE_ATTR",
    "PIPESTAT_NAMESPACE_ATTR_KEY",
    "PIPESTAT_CONFIG_ATTR_KEY",
    "PIPESTAT_RESULTS_FILE_ATTR_KEY",
    "LOOPER_GENERIC_PIPELINE",
    "PROJECT_PL_ARG",
    "SAMPLE_PL_ARG",
    "JOB_NAME_KEY",
    "PIPELINE_INTERFACE_PIPELINE_NAME_KEY",
    "PEP_CONFIG_KEY",
    "PEP_CONFIG_FILE_KEY",
    "COMPUTE_SETTINGS_VARNAME",
    "DEFAULT_COMPUTE_RESOURCES_NAME",
    "NEW_COMPUTE_KEY",
    "DEFAULT_CONFIG_FILEPATH",
    "DEFAULT_CONFIG_SCHEMA",
    "DEFAULT_COMPUTE_RESOURCES_NAME",
    "MESSAGE_BY_SUBCOMMAND",
    "SAMPLE_SELECTION_ATTRIBUTE_OPTNAME",
    "SAMPLE_EXCLUSION_OPTNAME",
    "SAMPLE_INCLUSION_OPTNAME",
    "SAMPLE_SELECTION_FLAG_OPTNAME",
    "SAMPLE_EXCLUSION_FLAG_OPTNAME",
    "DEBUG_JOBS",
    "DEBUG_COMMANDS",
    "DEBUG_EIDO_VALIDATION",
    "LOOPER_GENERIC_OUTPUT_SCHEMA",
    "LOOPER_GENERIC_COUNT_LINES",
]

FLAGS = ["completed", "running", "failed", "waiting", "partial"]
# TODO: flag info should come from pipestat status schema
APPEARANCE_BY_FLAG = {
    "completed": {"button_class": "{type}-success", "flag": "Completed"},
    "running": {"button_class": "{type}-primary", "flag": "Running"},
    "failed": {"button_class": "{type}-danger", "flag": "Failed"},
    "parital": {"button_class": "{type}-warning", "flag": "Partial"},
    "waiting": {"button_class": "{type}-info", "flag": "Waiting"},
}


def _get_apperance_dict(type, templ=APPEARANCE_BY_FLAG):
    """
    Based on the type of the HTML element provided construct the appearence
     mapping using the template

    :param dict templ: appearance template to populate
    :param str type: type of HTML element to populate template with
    :return dict: populated appearance template
    """
    from copy import deepcopy

    ret = deepcopy(templ)
    for flag, app_dict in ret.items():
        for key, app in app_dict.items():
            ret[flag][key] = ret[flag][key].format(type=type)
    return ret


# Debug keys
DEBUG_JOBS = "Jobs submitted"
DEBUG_COMMANDS = "Commands submitted"
DEBUG_EIDO_VALIDATION = "EidoValidationError"

# Compute-related (for divvy)
COMPUTE_SETTINGS_VARNAME = ["DIVCFG"]
DEFAULT_COMPUTE_RESOURCES_NAME = "default"
OLD_COMPUTE_KEY = "compute"
NEW_COMPUTE_KEY = "compute_packages"
DEFAULT_CONFIG_FILEPATH = os.path.join(
    os.path.dirname(__file__), "default_config", "divvy_config.yaml"
)
DEFAULT_CONFIG_SCHEMA = os.path.join(
    os.path.dirname(__file__), "schemas", "divvy_config_schema.yaml"
)
PRE_SUBMIT_HOOK_KEY = "pre_submit"
PRE_SUBMIT_PY_FUN_KEY = "python_functions"
PRE_SUBMIT_CMD_KEY = "command_templates"

LOGGING_LEVEL = "INFO"
CFG_ENV_VARS = ["LOOPER"]
TABLE_APPEARANCE_BY_FLAG = _get_apperance_dict("table")
BUTTON_APPEARANCE_BY_FLAG = _get_apperance_dict("btn btn")
NO_DATA_PLACEHOLDER = "NA"
PIFACE_KEY_SELECTOR = "pipeline_interfaces_key"
PIPELINE_INTERFACES_KEY = "pipeline_interfaces"
RESOURCES_KEY = "resources"
SAMPLE_PL_KEY = "sample_pipeline"
PROJECT_PL_KEY = "project_pipeline"
PIFACE_SCHEMA_SRC = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "schemas",
    "pipeline_interface_schema_{}.yaml",
)
EXTRA_SAMPLE_CMD_TEMPLATE = (
    "{%- if sample.command_extra is defined %} {sample.command_extra}   {% endif -%}"
)
EXTRA_PROJECT_CMD_TEMPLATE = (
    "{%- if looper.command_extra is defined %} {looper.command_extra}{% endif -%}"
)
DOTFILE_CFG_PTH_KEY = "config_file_path"
INPUT_SCHEMA_KEY = "input_schema"
OUTPUT_SCHEMA_KEY = "output_schema"
OBJECT_TYPES = ["object", "file", "image", "array"]
SAMPLE_YAML_PATH_KEY = "sample_yaml_path"
SAMPLE_YAML_PRJ_PATH_KEY = "sample_yaml_prj_path"
SUBMISSION_YAML_PATH_KEY = "submission_yaml_path"
SAMPLE_CWL_YAML_PATH_KEY = "sample_cwl_yaml_path"
SAMPLE_TOGGLE_ATTR = "toggle"
OUTKEY = "outputs"
JOB_NAME_KEY = "job_name"
VAR_TEMPL_KEY = "var_templates"
COMPUTE_KEY = "compute"
COMPUTE_PACKAGE_KEY = "package"
SIZE_DEP_VARS_KEY = "size_dependent_variables"
DYN_VARS_KEY = "dynamic_variables_command_template"
TEMPLATES_DIRNAME = "jinja_templates"
NOT_SUB_MSG = "> Not submitted: {}"
IMAGE_EXTS = (".png", ".jpg", ".jpeg", ".svg", ".gif")
# this strongly depends on pypiper's profile.tsv format
PROFILE_COLNAMES = ["pid", "hash", "cid", "runtime", "mem", "cmd", "lock"]

PIPELINE_INTERFACE_PIPELINE_NAME_KEY = "pipeline_name"

# pipestat configuration keys
DEFAULT_PIPESTAT_CONFIG_ATTR = "pipestat_config"
DEFAULT_PIPESTAT_RESULTS_FILE_ATTR = "pipestat_results_file"
PIPESTAT_NAMESPACE_ATTR_KEY = "namespace_attribute"
PIPESTAT_CONFIG_ATTR_KEY = "config_attribute"
PIPESTAT_RESULTS_FILE_ATTR_KEY = "results_file_path"

PIPE_ARGS_SECTION = "pipeline_args"
CLI_KEY = "cli"
LOOPER_KEY = "looper"
PIPESTAT_KEY = "pipestat"
NAMESPACE_ATTR_KEY = "namespace_attribute"
OUTDIR_KEY = "output_dir"
PEP_CONFIG_KEY = "pep_config"
PEP_CONFIG_FILE_KEY = "config_file"

RESULTS_SUBDIR_KEY = "results_subdir"
SUBMISSION_SUBDIR_KEY = "submission_subdir"
DRY_RUN_KEY = "dry_run"
FILE_CHECKS_KEY = "skip_file_checks"
EXAMPLE_COMPUTE_SPEC_FMT = "k1=v1 k2=v2"
SUBMISSION_FAILURE_MESSAGE = "Cluster resource failure"
LOOPER_DOTFILE_NAME = "." + LOOPER_KEY + ".yaml"
LOOPER_GENERIC_PIPELINE = "pipeline_interface.yaml"
LOOPER_GENERIC_OUTPUT_SCHEMA = "output_schema.yaml"
LOOPER_GENERIC_COUNT_LINES = "count_lines.sh"
POSITIONAL = [PEP_CONFIG_FILE_KEY, "command"]
SELECTED_COMPUTE_PKG = "package"
EXTRA_KEY = "_cli_extra"
ALL_SUBCMD_KEY = "all"
SAMPLE_PL_ARG = "sample_pipeline_interfaces"
PROJECT_PL_ARG = "project_pipeline_interfaces"


DEFAULT_CFG_PATH = os.path.join(os.getcwd(), LOOPER_DOTFILE_NAME)
CLI_PROJ_ATTRS = [
    OUTDIR_KEY,
    SUBMISSION_SUBDIR_KEY,
    PIPELINE_INTERFACES_KEY,
    RESULTS_SUBDIR_KEY,
    PIFACE_KEY_SELECTOR,
    COMPUTE_PACKAGE_KEY,
    DRY_RUN_KEY,
    FILE_CHECKS_KEY,
    SAMPLE_PL_ARG,
    PIPESTAT_KEY,
    DEFAULT_PIPESTAT_CONFIG_ATTR,
    PEP_CONFIG_KEY,
]

# resource package TSV-related consts
ID_COLNAME = "id"
FILE_SIZE_COLNAME = "max_file_size"
IMAGE_EXTS = (".png", ".jpg", ".jpeg", ".svg", ".gif")
# this strongly depends on pypiper's profile.tsv format
PROFILE_COLNAMES = ["pid", "hash", "cid", "runtime", "mem", "cmd", "lock"]


# Argument option names

SAMPLE_SELECTION_ATTRIBUTE_OPTNAME = "sel-attr"
SAMPLE_EXCLUSION_OPTNAME = "sel-excl"
SAMPLE_INCLUSION_OPTNAME = "sel-incl"
SAMPLE_SELECTION_FLAG_OPTNAME = "sel-flag"
SAMPLE_EXCLUSION_FLAG_OPTNAME = "exc-flag"

MESSAGE_BY_SUBCOMMAND = {
    "run": "Run or submit sample jobs.",
    "rerun": "Resubmit sample jobs with failed flags.",
    "runp": "Run or submit project jobs.",
    "table": "Write summary stats table for project samples.",
    "report": "Create browsable HTML report of project results.",
    "destroy": "Remove output files of the project.",
    "check": "Check flag status of current runs.",
    "clean": "Run clean scripts of already processed jobs.",
    "inspect": "Print information about a project.",
    "init": "Initialize looper config file.",
    "init-piface": "Initialize generic pipeline interface.",
    "link": "Create directory of symlinks for reported results.",
}



""" Exceptions for specific looper issues. """

from abc import ABCMeta
from collections.abc import Iterable

__author__ = "Vince Reuter"
__email__ = "vreuter@virginia.edu"

_all__ = [
    "DuplicatePipelineKeyException",
    "InvalidResourceSpecificationException",
    "JobSubmissionException",
    "LooperError",
    "MissingPipelineConfigurationException",
    "PipelineInterfaceConfigError",
    "PipelineInterfaceRequirementsError",
    "MisconfigurationException",
]


class LooperError(Exception):
    """Base type for custom package errors."""

    __metaclass__ = ABCMeta


class SampleFailedException(LooperError):
    """An exception to be raised if any sample fails"""

    pass


class MisconfigurationException(LooperError):
    """Duplication of pipeline identifier precludes unique pipeline ref."""

    def __init__(self, key):
        super(MisconfigurationException, self).__init__(key)


class RegistryPathException(LooperError):
    """Duplication of pipeline identifier precludes unique pipeline ref."""

    def __init__(self, msg):
        super(RegistryPathException, self).__init__(msg)


class DuplicatePipelineKeyException(LooperError):
    """Duplication of pipeline identifier precludes unique pipeline ref."""

    def __init__(self, key):
        super(DuplicatePipelineKeyException, self).__init__(key)


class InvalidResourceSpecificationException(LooperError):
    """Pipeline interface resources--if present--needs default."""

    def __init__(self, reason):
        super(InvalidResourceSpecificationException, self).__init__(reason)


class JobSubmissionException(LooperError):
    """Error type for when job submission fails."""

    def __init__(self, sub_cmd, script):
        self.script = script
        reason = "Error for command {} and script '{}'".format(sub_cmd, self.script)
        super(JobSubmissionException, self).__init__(reason)


class PipestatConfigurationException(LooperError):
    """Error type for when command fails due to missing pipestat config"""

    def __init__(
        self,
        sub_cmd,
    ):
        reason = "Pipestat must be configured for command {}".format(sub_cmd)
        super(PipestatConfigurationException, self).__init__(reason)


class MissingPipelineConfigurationException(LooperError):
    """A selected pipeline needs configuration data."""

    def __init__(self, pipeline):
        super(MissingPipelineConfigurationException, self).__init__(pipeline)


class PipelineInterfaceConfigError(LooperError):
    """Error with PipelineInterface config data during construction."""

    def __init__(self, context):
        """
        For exception context, provide message or collection of missing sections.

        :param str | Iterable[str] context:
        """
        if not isinstance(context, str) and isinstance(context, Iterable):
            context = "Missing section(s): {}".format(", ".join(context))
        super(PipelineInterfaceConfigError, self).__init__(context)


class PipelineInterfaceRequirementsError(LooperError):
    """Invalid specification of pipeline requirements in interface config."""

    def __init__(self, typename_by_requirement):
        super(PipelineInterfaceRequirementsError, self).__init__(
            "{} invalid requirements: {}".format(
                len(typename_by_requirement), typename_by_requirement
            )
        )
        self.error_specs = typename_by_requirement
